import unittest
import torch as T
import torch.nn as nn
from lpd.metrics import BinaryAccuracy, BinaryAccuracyWithLogits, CategoricalAccuracy, CategoricalAccuracyWithLogits
from lpd.metrics import TopKCategoricalAccuracy, TruePositives, TrueNegatives, FalseNegatives, FalsePositives
from lpd.metrics.confusion_matrix import ConfusionMatrix

class TestMetrics(unittest.TestCase):

    def test_binary_accuracy(self):
        metric = BinaryAccuracy()

        y_pred = T.Tensor([0.,0.,0.,1.])
        y_true = T.Tensor([0.,0.,0.,1.])
        self.assertEqual(metric(y_pred, y_true), 1.0)

        y_pred = T.Tensor([1.,0.,0.,0.])
        y_true = T.Tensor([0.,0.,0.,1.])
        self.assertEqual(metric(y_pred, y_true), 0.5)

        y_pred = T.Tensor([0.,0.,0.,0])
        y_true = T.Tensor([1.,1.,1.,1.])
        self.assertEqual(metric(y_pred, y_true), 0.0)

        y_pred = T.Tensor([[0.,0.],[1.,0.]])
        y_true = T.Tensor([[1.,0.],[1.,0.]])
        self.assertEqual(metric(y_pred, y_true), 0.75)

    def test_binary_accuracy_with_logits(self):
        metric = BinaryAccuracyWithLogits()

        y_pred = T.Tensor([-0.3,-0.3,-0.3,1.7])
        y_true = T.Tensor([0.0,0.0,0.0,1.0])
        self.assertEqual(metric(y_pred, y_true), 1.0)

        y_pred = T.Tensor([1.7,-0.3,-0.3,-0.3])
        y_true = T.Tensor([0.0,0.0,0.0,1.0])
        self.assertEqual(metric(y_pred, y_true), 0.5)

        y_pred = T.Tensor([-0.3,-0.3,-0.3,-0.3])
        y_true = T.Tensor([1.0,1.0,1.0,1.0])
        self.assertEqual(metric(y_pred, y_true), 0.0)

        y_pred = T.Tensor([[-0.3,-0.3],[1.7,-0.3]])
        y_true = T.Tensor([[1.0,0.0],[1.0,0.0]])
        self.assertEqual(metric(y_pred, y_true), 0.75)

    def test_categorical_accuracy(self):
        dim = 4
        metric = CategoricalAccuracy()

        # y_pred:
        # [1,0,0,0]
        # [0,1,0,0]
        # [0,0,1,0]
        # [0,0,0,1]

        # y_true:
        # [0,1,2,3]
        y_pred = T.eye(dim)
        y_true = T.LongTensor(list(range(dim)))
        self.assertEqual(metric(y_pred, y_true), 1.0)
        
        # y_pred:
        # [1,0,0,0]
        # [1,0,0,0]
        # [1,0,0,0]
        # [1,0,0,0]

        # y_true:
        # [0,1,2,3]
        y_pred = T.zeros(dim, dim)
        y_pred.data[:, 0] = T.ones(dim)
        y_true = T.LongTensor(list(range(dim)))
        self.assertEqual(metric(y_pred, y_true), 0.25)

    def test_categorical_accuracy_with_logits(self):
        dim = 4
        metric = CategoricalAccuracyWithLogits()

        y_pred = T.Tensor([[5,1,1],[2,9,1],[2,3,6]])
        y_true = T.Tensor([0,1,2])
        self.assertEqual(metric(y_pred, y_true), 1.0)

        y_pred = T.Tensor([[5,1,1],[2,9,1],[5,1,1]])
        y_true = T.Tensor([0,1,2])
        self.assertEqual(metric(y_pred, y_true), 2./3.)

        y_pred = T.Tensor([[5,1,1],[2,1,9],[5,1,1]])
        y_true = T.Tensor([0,1,2])
        self.assertEqual(metric(y_pred, y_true), 1./3.)

        y_pred = T.Tensor([[2,5,1],[2,1,9],[5,1,1]])
        y_true = T.Tensor([0,1,2])
        self.assertEqual(metric(y_pred, y_true), 0.0)

    def test_top_k_categorical_accuracy(self):
        # K = 1
        metric = TopKCategoricalAccuracy(k=1)

        y_pred = T.Tensor([[0.1, 0.9, 0.8], [0.05, 0.95, 0], [0.05, 0.95, 0]])
        y_true = T.Tensor([1,1,1])
        self.assertEqual(metric(y_pred, y_true), 1.0)

        y_pred = T.Tensor([[0.1, 0.9, 0.5], [0.05, 0.8, 0.]])
        y_true = T.Tensor([2,1])
        self.assertEqual(metric(y_pred, y_true), 0.5)

        y_pred = T.Tensor([[0.1, 0.3, 0.8], [0.05, 0.2, 0.3], [0.8, 0.4, 0.3]])
        y_true = T.Tensor([1,1,0])
        self.assertEqual(metric(y_pred, y_true), 1/3)

        y_pred = T.Tensor([[0.1, 0.3, 0.8], [0.05, 0.2, 0.3], [0.8, 0.4, 0.3], [0.8, 0.4, 0.3]])
        y_true = T.Tensor([1,1,0,2])
        self.assertEqual(metric(y_pred, y_true), 0.25)


        # K = 2
        metric = TopKCategoricalAccuracy(k=2)

        y_pred = T.Tensor([[0.1, 0.9, 0.8], [0.05, 0.95, 0], [0.05, 0.95, 0]])
        y_true = T.Tensor([1,1,1])
        self.assertEqual(metric(y_pred, y_true), 1.0)

        y_pred = T.Tensor([[0.1, 0.9, 0.01], [0.05, 0.8, 0.]])
        y_true = T.Tensor([2,1])
        self.assertEqual(metric(y_pred, y_true), 0.5)

        y_pred = T.Tensor([[0.1, 0.3, 0.8], [0.05, 0.2, 0.3], [0.3, 0.4, 0.8]])
        y_true = T.Tensor([1,1,0])
        self.assertEqual(metric(y_pred, y_true), 2/3)

        y_pred = T.Tensor([[0.3, 0.1, 0.8], [0.5, 0.2, 0.3], [0.8, 0.4, 0.3], [0.8, 0.4, 0.3]])
        y_true = T.Tensor([1,1,0,2])
        self.assertEqual(metric(y_pred, y_true), 0.25)

    def test_confusion_matrix(self):
        # WE TEST ALL KINDS OF CLASSIFICATION OUTPUTS, THERE MIGHT BE MORE, 
        # THATS WHY THERE IS THE predictions_to_classes_convertor Parameter
        # WHERE YOU CAN PASS YOUR OWN FUNCTION TO CONVERT y_pred TO CLASS INDICES

        #-------------------- BINARY SINGE DIGIT MODE - LEN(SHAPE) = 1 --------------------#
        # DEFINE FUNCTIONS TO CONVERT y_pred TO CLASSES

        labels = ["a", "b"]
        num_classes = len(labels)
        actual    = T.Tensor([1,1,1,1,0,0,0,0])
        predicted = T.Tensor([1,1,1,1,0,0,0,1])

        metric = ConfusionMatrix(num_classes, labels=labels)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[3,0],[1,4]])).all())



        #-------------------- BINARY SINGE DIGIT MODE WITH LOGITS - LEN(SHAPE) = 1 --------------------#

        labels = ["a", "b"]
        num_classes = len(labels)
        actual    = T.Tensor([1,1,1,1,0,0,0,0])
        predicted = T.Tensor([1,1,1,1,-1,-1,-1,1])

        metric = ConfusionMatrix(num_classes, labels=labels, threshold=0)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[3,0],[1,4]])).all())


        #-------------------- BINARY SINGE DIGIT MODE - LEN(SHAPE) = 2 --------------------#

        labels = ["a", "b"]
        num_classes = len(labels)
        actual    = T.Tensor([0,0,0,0,1,1,1,1])
        predicted = T.Tensor([[0],[0],[0],[0],[1],[1],[1],[0]])

        metric = ConfusionMatrix(num_classes, labels=labels)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[4,1],[0,3]])).all())


        #-------------------- CLASSIFICATION 2 CLASSES --------------------#

        labels = ["a", "b"]
        num_classes = len(labels)
        actual    = T.Tensor([0,0,0,0,1,1,1,1])
        predicted = T.Tensor([[1,0],[1,0],[1,0],[1,0],[0,1],[0,1],[0,1],[1,0]])

        metric = ConfusionMatrix(num_classes, labels=labels)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[4,1],[0,3]])).all())


        #-------------------- CLASSIFICATION 3 CLASSES --------------------#

        labels = ["a", "b", "z"]
        num_classes = len(labels)
        emb = nn.Embedding(num_classes,num_classes)
        emb.weight.data = T.eye(num_classes)
        actual    = T.Tensor([0,0,0,0,1,1,1,1])
        predicted = T.Tensor([0,0,0,0,1,1,1,2]).long()
        predicted = emb(predicted)

        metric = ConfusionMatrix(num_classes, labels=labels)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[4,0,0],[0,3,0],[0,1,0]])).all())


        #-------------------- CLASSIFICATION 3 CLASSES --------------------#

        labels = ["a", "b", "c"]
        num_classes = len(labels)
        emb = nn.Embedding(num_classes,num_classes)
        emb.weight.data = T.eye(num_classes)
        actual      = T.Tensor([0, 1, 2, 2, 1, 2, 2, 1, 0, 0, 1, 0, 1, 2, 0, 1, 2])
        predicted   = T.Tensor([0, 1, 1, 2, 0, 2, 0, 1, 2, 0, 1, 1, 1, 2, 0, 0, 2]).long()
        predicted  = emb(predicted)

        metric = ConfusionMatrix(num_classes, labels=labels)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[3,2,1],[1,4,1],[1,0,4]])).all())


        #-------------------- CLASSIFICATION 4 CLASSES --------------------#
        labels = ["a", "b", "z", "x"]
        num_classes = len(labels)
        emb = nn.Embedding(num_classes,num_classes)
        emb.weight.data = T.eye(num_classes)
        actual    = T.Tensor([0,0,0,2,2,1,1,1])
        predicted = T.Tensor([0,0,0,0,1,1,1,3]).long()
        predicted  = emb(predicted)

        metric = ConfusionMatrix(num_classes, labels=labels)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[3,0,1,0],[0,2,1,0],[0,0,0,0],[0,1,0,0]])).all())


        #-------------------- LOGITS - CLASSIFICATION 4 CLASSES --------------------#

        labels = ["a", "b", "z", "x"]
        num_classes = len(labels)
        emb = nn.Embedding(num_classes,num_classes)
        emb.weight.data = (T.eye(num_classes) * 0.6) + 0.2
        actual    = T.Tensor([0,0,0,2,2,1,1,1])
        predicted = T.Tensor([0,0,0,0,1,1,1,3]).long()
        predicted  = emb(predicted)

        metric = ConfusionMatrix(num_classes, labels=labels, threshold=0.75)
        metric.update_state(predicted, actual)

        conf = metric.get_confusion_matrix()
        self.assertTrue((conf==T.Tensor([[3,0,1,0],[0,2,1,0],[0,0,0,0],[0,1,0,0]])).all())

    def test_tp_tn_fp_fn(self):
        # BINARY
        labels = ["a", "b"]
        num_classes = len(labels)
        # 4 TP, 3 TN, 5 FP, 2 FN
        actual    = T.Tensor([1,1,1,1,0,0,0,0,0,0,0,0,1,1])
        predicted = T.Tensor([1,1,1,1,0,0,0,1,1,1,1,1,0,0])

        confusion_matrix = ConfusionMatrix(num_classes=num_classes, labels=labels)
        confusion_matrix.update_state(predicted, actual)

        metric = TruePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([4])).all())


        metric = TrueNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([3])).all())


        metric = FalsePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([5])).all())


        metric = FalseNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([2])).all())


        # BINARY AS 2 CLASSES
        labels = ["a", "b"]
        num_classes = len(labels)
        emb = nn.Embedding(num_classes,num_classes)
        emb.weight.data = T.eye(num_classes)
        # 4 TP, 3 TN, 5 FP, 2 FN
        actual    = T.Tensor([1,1,1,1,0,0,0,0,0,0,0,0,1,1])
        predicted = T.Tensor([1,1,1,1,0,0,0,1,1,1,1,1,0,0]).long()
        predicted  = emb(predicted)

        confusion_matrix = ConfusionMatrix(num_classes=num_classes, labels=labels)
        confusion_matrix.update_state(predicted, actual)

        metric = TruePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([4])).all())


        metric = TrueNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([3])).all())


        metric = FalsePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([5])).all())


        metric = FalseNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([2])).all())



        # 4 classes

        labels = ["a", "b", "c", "d"]
        num_classes = len(labels)
        emb = nn.Embedding(num_classes,num_classes)
        emb.weight.data = T.eye(num_classes)
        actual    = T.Tensor([0,0,0,2,2,1,1,1])
        predicted = T.Tensor([0,0,0,0,1,1,1,3]).long()
        predicted  = emb(predicted)

        confusion_matrix = ConfusionMatrix(num_classes, labels=labels)
        confusion_matrix.update_state(predicted, actual)

        # [[3,0,1,0],
        #  [0,2,1,0],
        #  [0,0,0,0],
        #  [0,1,0,0]]

        metric = TruePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([3,2,0,0])).all())

        metric = FalsePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([1,1,0,1])).all())

        metric = FalseNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([0,1,2,0])).all())

        metric = TrueNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([4,4,6,7])).all())



        # 4 CLASSES MULTI-LABEL

        labels = ["a", "b", "c", "d"]
        num_classes = len(labels)
        emb = nn.Embedding(num_classes,num_classes)
        emb.weight.data = T.Tensor([[1,0,1,1],[0,1,0,0],[0,0,1,1],[0,0,0,1]])
        actual    = T.Tensor([0,0,0,2,2,1,1,1])
        predicted = T.Tensor([0,0,0,0,1,1,1,3]).long()
        predicted  = emb(predicted)

        def multilabel_ypred_to_indices(y_pred, t_true):
            # FOR SIMPLICITY, THE FIRST "ARGMAX" INDEX FOUND, IS CONSIDERED THE CHOSEN CLASS
            classes = T.max(y_pred, dim=1)[1]
            return classes

        confusion_matrix = ConfusionMatrix(num_classes, labels=labels, predictions_to_classes_convertor=multilabel_ypred_to_indices)
        confusion_matrix.update_state(predicted, actual)

        metric = TruePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([3,2,0,0])).all())

        metric = FalsePositives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([1,1,0,1])).all())

        metric = FalseNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([0,1,2,0])).all())

        metric = TrueNegatives(num_classes)
        metric._set_confusion_matrix(confusion_matrix) #IN REAL-TIME THIS IS BEING HANDLED BY TRAINER-STATS
        result = metric(predicted, actual)
        self.assertTrue((result==T.Tensor([4,4,6,7])).all())




from lpd.enums.phase import Phase
from lpd.enums.monitor_mode import MonitorMode
from lpd.enums.monitor_type import MonitorType
from lpd.enums.state import State
from lpd.enums.stats_type import StatsType
from lpd.enums.confusion_matrix_based_metric import ConfusionMatrixBasedMetric
from lpd.enums.metric_method import MetricMethod


import torch.optim as optim
import torch.nn as nn

from lpd.trainer import Trainer
from lpd.callbacks import StatsPrint, SchedulerStep, LossOptimizerHandler
from lpd.enums import Phase, State 
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu


def get_parameters():
    # N is batch size; D_in is input dimension;
    # H is hidden dimension; D_out is output dimension.
    N, D_in, H, D_out = 64, 1000, 100, 10
    num_epochs = 10
    data_loader = eu.examples_data_generator(N, D_in, D_out)
    data_loader_steps = 100
    return N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps

def get_trainer(N, D_in, H, D_out, data_loader, data_loader_steps):

    device = tu.get_gpu_device_if_available()

    model = eu.get_basic_model(D_in, H, D_out).to(device)

    loss_func = nn.MSELoss(reduction='sum').to(device)
   
    optimizer = optim.Adam(model.parameters(), lr=1e-4)

    # LETS ADD A StepLR SCHEDULER 
    scheduler = optim.lr_scheduler.StepLR(optimizer=optimizer, gamma=0.999, step_size=1)
    
    metric_name_to_func = None # THIS EXAMPLE DOES NOT USE METRICS, ONLY LOSS

    # LETS ADD SchedulerStep WITH apply_on_phase=Phase.BATCH_END
    # AND apply_on_states=State.TRAIN
    # IT MEANS THAT THE SchedulerStep WILL BE INVOKED AT THE END OF EVERY BATCH, BUT, WILL ONLY BE APPLIED WHEN 
    # IN TRAIN MODE, AND WILL BE IGNORED IN VAL/TEST MODES
    # NOTICE!!! WE USE verbose=1 TO SEE THE PRINTS FOR THIS EXAMPLE, BUT YOU MIGHT PREFER TO USE verbose=0 or verbose=2
    # BECAUSE ON BATCH LEVEL IT WILL PRINT A LOT 
    callbacks = [   
                    LossOptimizerHandler(),
                    SchedulerStep(apply_on_phase=Phase.BATCH_END, apply_on_states=State.TRAIN, verbose=1), #CAN ALSO BE IN FORM OF ARRAY - apply_on_states=[State.TRAIN]
                    StatsPrint(apply_on_phase=Phase.EPOCH_END)
                ]

    trainer = Trainer(model=model, 
                      device=device, 
                      loss_func=loss_func, 
                      optimizer=optimizer,
                      scheduler=scheduler,
                      metric_name_to_func=metric_name_to_func, 
                      train_data_loader=data_loader, 
                      val_data_loader=data_loader,
                      train_steps=data_loader_steps,
                      val_steps=data_loader_steps,
                      callbacks=callbacks,
                      name='Scheduler-Step-On-Batch-Example')
    return trainer

def run():
    gu.seed_all(42)  # BECAUSE ITS THE ANSWER TO LIFE AND THE UNIVERSE

    N, D_in, H, D_out, num_epochs, data_loader, data_loader_steps = get_parameters()

    trainer = get_trainer(N, D_in, H, D_out, data_loader, data_loader_steps)
    
    trainer.summary()

    #WE NOW EXPECT SchedulerStep TO INVOKE StepLR AT THE END OF EVERY BATCH
    trainer.train(num_epochs)

    trainer.evaluate(data_loader, data_loader_steps)


import os
from setuptools import setup, find_packages
from io import open as io_open

src_dir = os.path.abspath(os.path.dirname(__file__))

install_requires = []
requirements_dev = os.path.join(src_dir, 'requirements-dev.txt')
with io_open(requirements_dev, mode='r') as fd:
    install_requires = [i.strip().split('#', 1)[0].strip() for i in fd.read().strip().split('\n')]

README_md = ''
fndoc = os.path.join(src_dir, 'README.md')
with io_open(fndoc, mode='r', encoding='utf-8') as fd:
    README_md = fd.read()

setup(
    name='lpd',
    version='0.3.4',
    description='A Fast, Flexible Trainer with Callbacks and Extensions for PyTorch',
    long_description_content_type='text/markdown',
    long_description=README_md,
    license='MIT Licences',
    url='https://github.com/roysadaka/lpd',
    author='Roy Sadaka',
    maintainer='lpd developers',
    maintainer_email='torch.lpd@gmail.com',
    packages=find_packages(),
    install_requires=install_requires,
    python_requires='>=3.6',
    classifiers=[
        # (https://pypi.org/pypi?%3Aaction=list_classifiers)
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Utilities'
    ],
    keywords=['pytorch,trainer,callback,callbacks,earlystopping,tensorboard,modelcheckpoint,checkpoint,layers,dense,metrics,predictor,binary accuracy,extensions,track,monitor,machine,deep learning,neural,networks,AI,keras decay,confusion matrix']
)


import os
import unittest

import torch.optim as optim
import torch.nn as nn
from lpd.trainer import Trainer
from lpd.callbacks import StatsPrint, SchedulerStep, LossOptimizerHandler, ModelCheckPoint, CallbackMonitor
from lpd.extensions.custom_schedulers import KerasDecay
from lpd.enums import Phase, State, MonitorType, StatsType, MonitorMode
from lpd.metrics import BinaryAccuracyWithLogits, CategoricalAccuracyWithLogits
import lpd.utils.torch_utils as tu
import lpd.utils.general_utils as gu
import examples.utils as eu


class TestCallbacks(unittest.TestCase):
    def test_stats_print_validations(self):
        # ASSERT INVALID INIT
        self.assertRaises(ValueError, StatsPrint, train_metrics_monitors=CallbackMonitor(patience=None, 
                                                                                 monitor_type=MonitorType.METRIC,
                                                                                 stats_type=StatsType.VAL,
                                                                                 monitor_mode=MonitorMode.MAX,
                                                                                 metric_name='Accuracy'))
        # ASSERT VALID INIT
        StatsPrint(train_metrics_monitors=CallbackMonitor(patience=None, 
                                                            monitor_type=MonitorType.METRIC,
                                                            stats_type=StatsType.TRAIN,
                                                            monitor_mode=MonitorMode.MAX,
                                                            metric_name='Accuracy'))


import torch as T
import torch.nn as nn
import math

class MatMul2D(nn.Module):
    def __init__(self, transpose_b, name=None):
        super(MatMul2D, self).__init__()
        #PARAMS
        self.transpose_b = transpose_b
        self.name = name if name else 'mat_mul'

    def forward(self, a, b):
        if self.transpose_b:
            return T.bmm(a, b.transpose(1,2))
        return T.bmm(a, b)

class Dense(nn.Module):
    def __init__(self, in_dim, out_dim, use_bias=True, activation=None, name=None):
        super(Dense, self).__init__()
        #PARAMS
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.use_bias = use_bias
        self.activation = activation
        self.name = name if name else 'dense'
        #LAYERS
        self.fc = nn.Linear(self.in_dim, self.out_dim, bias=self.use_bias)
        self._init_weights()

    def _init_weights(self):
        nn.init.xavier_uniform_(self.fc.weight)
        if self.use_bias:
            nn.init.zeros_(self.fc.bias)

    def forward(self, inputs):
        x = self.fc(inputs)
        if self.activation:
            x = self.activation(x)
        return x

class Attention(nn.Module):
    """
        The architecture is based on the paper “Attention Is All You Need”
        Usage (1)
        It can be used as Attention in transformer if q,k,v share the same dimensions.
        
        Usage (2)
        It can also be used as a method to aggregate a group of vectors into 1 vector if q dimensions are (batch, 1, key_dim)
        that way, instead of using Sum, or Average, you can have a learnable query vector (or a few of them) that will learn the aggregation function.
        See example in lpd.examples.multiple_inputs.model, where we define external_query_attention like so:
        external_query_attention = Attention(key_dim=config.EMBEDDINGS_SIZE, use_query_dense=True)  

        Args:
        key_dim - as defined in the paper, the number of expected features in the encoder inputs
        use_query_dense - whether to pass q input into another Dense layer, mostly used in Usage (2), to
                          run q into a transformation that will transform it into the vector space of k and v
        name - optional, any string to describe this layer
    """
    def __init__(self, key_dim, use_query_dense=False, name=None):
        super(Attention, self).__init__()
        #PARAMS
        self.key_dim            = key_dim
        self.sqrt_key_dim       = key_dim ** 0.5
        self.use_query_dense    = use_query_dense
        self.name               = name if name else 'attention'
        #LAYERS
        self.mat_mul2d          = MatMul2D(transpose_b=False, name = f'{self.name}__MatMul2D')
        self.mat_mul2d_t        = MatMul2D(transpose_b=True, name = f'{self.name}__MatMul2DT')
        self.softmax_last_dim   = nn.Softmax(dim=-1)
        if self.use_query_dense:
            # SOMETIMES WE WANT TO GO THROUGH ANOTHER TRANSFORMATION BEFORE RUNNING THE QUERY,
            # FOR EXAMPLE, WHEN THIS IS USED AS A STANDALONE LAYER
            self.query_dense    = Dense(in_dim=self.key_dim, out_dim=self.key_dim, use_bias=False, activation=None, name = f'{self.name}__Dense')

    def forward(self, q,k,v, mask = None):
        # q:    (batch, ?, key_dim)             where "?" can be 1 or seq_len
        # k:    (batch, seq_len, key_dim)
        # v:    (batch, seq_len, key_dim)
        # mask: (batch, 1, seq_len)

        # APPLY ATTENTION:
        #                       (     Q * Kt     )
        #               softmax (   ----------   ) * V
        #                       (    sqrt(dk)    )

        if self.use_query_dense:
            q = self.query_dense(q)                                            # (batch, seq_len, key_dim)

        q_k = self.mat_mul2d_t(q, k)                                           # (batch, ?, seq_len)
        scores = q_k / self.sqrt_key_dim                                       # (batch, ?, seq_len)

        if mask is not None:
            mask_ready = T.log(mask)                                           # (batch, 1, seq_len)
            scores += mask_ready                                               # (batch, ?, seq_len) (+= is doing broadcasting)

        attention_weights = self.softmax_last_dim(scores)                      # (batch, ?, seq_len)
        attention_output = self.mat_mul2d(attention_weights, v)                # (batch, ?, key_dim)

        return attention_output                                                # (batch, ?, key_dim)

class AttentionHead(nn.Module):
    def __init__(self, in_dim, key_dim, name=None):
        super(AttentionHead, self).__init__()
        #PARAMS
        self.in_dim = in_dim
        self.key_dim = key_dim
        self.sqrt_key_dim = key_dim ** 0.5
        self.name = name if name else 'Attention-Head'

        #LAYERS
        self.query_dense  = Dense(self.in_dim, self.key_dim, use_bias=True, activation=None, name = f'{self.name}__Q-Dense')
        self.key_dense    = Dense(self.in_dim, self.key_dim, use_bias=True, activation=None, name = f'{self.name}__K-Dense')
        self.value_dense  = Dense(self.in_dim, self.key_dim, use_bias=True, activation=None, name = f'{self.name}__V-Dense')
        self.att          = Attention(self.key_dim, name = f'{self.name}__Attention')

    def forward(self, inputs, mask = None):     # inputs:(batch, seq_len, emb_size), mask:(batch, seq_len)
        q = self.query_dense(inputs)            # (batch, seq_len, key_dim)
        k = self.key_dense(inputs)              # (batch, seq_len, key_dim)
        v = self.value_dense(inputs)            # (batch, seq_len, key_dim)
        return self.att(q,k,v,mask)             # (batch, seq_len, key_dim)

class MultiHeadAttention(nn.Module):
    def __init__(self, in_dim,
                       key_dim,
                       out_dim,
                       num_heads,
                       drop_out_proba,
                       name=None):
        super(MultiHeadAttention, self).__init__()
        #PARAMS
        self.in_dim = in_dim
        self.key_dim = key_dim
        self.out_dim = out_dim
        self.num_heads = num_heads
        self.drop_out_proba = drop_out_proba
        self.name = name if name else 'Multi-Head-Attention'

        #LAYERS
        self.attention_heads = nn.ModuleList([AttentionHead(self.in_dim, self.key_dim, name = f'{self.name}__H{i}') for i in range(self.num_heads)])
        self.output_dense = Dense(in_dim=self.num_heads*self.key_dim ,out_dim=self.out_dim, use_bias=True, activation=None, name = f'{self.name}__Out-Dense')
        self.dropout = nn.Dropout(p=self.drop_out_proba)
        self.norm = nn.LayerNorm(normalized_shape=self.out_dim) # WILL APPLY NORM OVER THE LAST DIMENTION ONLY

    def forward(self, inputs, mask=None):                                                     # inputs.shape: (batch, seq_len, emb_size == out_dim)
        attention_outputs = [head(inputs, mask = mask) for head in self.attention_heads]      # [ (batch, seq_len, key_dim) ]
        concatenated = T.cat(attention_outputs, dim=-1)                                       # (batch, seq_len, key_dim * num_heads)
        output = self.output_dense(concatenated)                                              # (batch, seq_len, out_dim)
        output = self.dropout(output)                                                         # (batch, seq_len, out_dim)
        return self.norm(inputs + output)                             # RESIDUAL & NORM       # (batch, seq_len, out_dim)

class TransformerEncoderFeedForward(nn.Module):
    def __init__(self, in_dim,
                       out_dim,
                       drop_out_proba,
                       expansion_rate,
                       name=None):
        super(TransformerEncoderFeedForward, self).__init__()
        #PARAMS
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.drop_out_proba = drop_out_proba
        self.expansion_rate = expansion_rate
        self.name = name if name else 'Transformer-Encoder__Feed-Forward'

        #LAYERS
        self.hidden_dense = Dense(in_dim=self.in_dim, out_dim=self.out_dim * self.expansion_rate, use_bias=True, activation=nn.ReLU(), name = f'{self.name}__Hidden-Dense')
        self.output_dense = Dense(in_dim=self.out_dim * self.expansion_rate, out_dim=self.out_dim, use_bias=True, activation=None, name = f'{self.name}__Out-Dense')
        self.dropout = nn.Dropout(p=self.drop_out_proba)

        self.norm = nn.LayerNorm(normalized_shape=self.out_dim)   # WILL APPLY NORM OVER THE LAST DIMENSION ONLY

    def forward(self, inputs):                                              # (batch, seq_len, out_dim)
        hidden_values = self.hidden_dense(inputs)                           # (batch, seq_len, out_dim * expansion_rate)
        output = self.output_dense(hidden_values)                           # (batch, seq_len, out_dim)
        output = self.dropout(output)                                       # (batch, num_elements, out_dim)
        return self.norm(inputs + output)    #RESIDUAL & NORM               # (batch, num_elements, out_dim)

class TransformerBlock(nn.Module):
    def __init__(self, in_dim,
                       key_dim,
                       out_dim,
                       num_heads,
                       drop_out_proba,
                       ff_expansion_rate,
                       name = None):
        super(TransformerBlock, self).__init__()
        #PARAMS
        self.in_dim = in_dim
        self.key_dim = key_dim
        self.out_dim = out_dim
        self.num_heads = num_heads
        self.drop_out_proba = drop_out_proba
        self.ff_expansion_rate = ff_expansion_rate
        self.name = name if name else 'Transformer-Encoder'
        #LAYERS
        self.multi_head_self_attention = MultiHeadAttention(self.in_dim,
                                                            self.key_dim,
                                                            self.out_dim,
                                                            self.num_heads,
                                                            self.drop_out_proba,
                                                            name = f'{self.name}__MH')

        self.feed_forward = TransformerEncoderFeedForward(self.out_dim,
                                                          self.out_dim,
                                                          self.drop_out_proba,
                                                          self.ff_expansion_rate,
                                                          name = f'{self.name}__FF')


    def forward(self, inputs, mask=None):                                           # inputs: (batch, num_elements, emb_size)
        attended = self.multi_head_self_attention(inputs=inputs, mask=mask)         # (batch, num_elements, out_dim)
        feed_forward = self.feed_forward(inputs=attended)                            # (batch, num_elements, out_dim)
        return feed_forward

class PositionalEncoding(nn.Module):
    #COPIED FROM https://pytorch.org/tutorials/beginner/transformer_tutorial.html
    def __init__(self, embedding_size, dropout_rate=0.1, maximum_position_encoding=5000):
        super(PositionalEncoding, self).__init__()
        #PARAMS
        self.embedding_size = embedding_size
        self.dropout_rate = dropout_rate
        self.maximum_position_encoding = maximum_position_encoding
        #LAYERS
        self.dropout = nn.Dropout(p=self.dropout_rate)
        self._create_positional_encoding()

    def _create_positional_encoding(self):
        if self.maximum_position_encoding:
            pe = T.zeros(self.maximum_position_encoding, self.embedding_size)
            position = T.arange(0, self.maximum_position_encoding, dtype=T.float).unsqueeze(1)
            div_term = T.exp(T.arange(0, self.embedding_size, 2).float() * (-math.log(10000.0) / self.embedding_size))
            pe[:, 0::2] = T.sin(position * div_term)
            pe[:, 1::2] = T.cos(position * div_term)
            pe = pe.unsqueeze(0).transpose(0, 1)
            self.register_buffer('pe', pe)

    def forward(self, inputs):
        inputs = inputs + self.pe[:inputs.size(0), :]
        return self.dropout(inputs)

class TransformerEncoderStack(nn.Module):
    def __init__(self, in_dim,
                       key_dim,
                       out_dim,
                       num_encoders,
                       num_heads,
                       drop_out_proba,
                       ff_expansion_rate,
                       maximum_position_encoding=None,
                       name=None):
        super(TransformerEncoderStack, self).__init__()
        #PARAMS
        self.in_dim = in_dim
        self.key_dim = key_dim
        self.out_dim = out_dim
        self.num_encoders = num_encoders
        self.num_heads = num_heads
        self.drop_out_proba = drop_out_proba
        self.ff_expansion_rate = ff_expansion_rate
        self.maximum_position_encoding = maximum_position_encoding
        self.name = name if name else 'Transformer-Encoder-Stack'

        #LAYERS
        self.transformer_blocks = nn.ModuleList([TransformerBlock(self.in_dim,
                                                    self.key_dim,
                                                    self.out_dim,
                                                    self.num_heads,
                                                    self.drop_out_proba,
                                                    self.ff_expansion_rate,
                                                    name=f'{self.name}__E{i}')
                                                    for i in range(self.num_encoders)])
        self.pos_encoder = PositionalEncoding(self.in_dim, self.drop_out_proba, self.maximum_position_encoding)

    def forward(self, inputs, mask=None):
        outputs = inputs                                        # (batch, seq_len, emb_size)

        #POSITION
        if self.maximum_position_encoding is not None:
            outputs = self.pos_encoder(outputs)

        for encoder_layer in self.transformer_blocks:
            outputs = encoder_layer(inputs=outputs, mask=mask)
        return outputs                                # (batch, seq_len, out_dim)   <-- USUALLY out_dim = emb_size








# class TransformerDecoderBlock(nn.Module):
#     def __init__(self, in_dim,
#                        key_dim,
#                        out_dim,
#                        num_heads,
#                        drop_out_proba,
#                        ff_expansion_rate,
#                        name = None):
#         super(TransformerDecoderBlock, self).__init__()
#         #PARAMS
#         self.in_dim = in_dim
#         self.key_dim = key_dim
#         self.out_dim = out_dim
#         self.num_heads = num_heads
#         self.drop_out_proba = drop_out_proba
#         self.ff_expansion_rate = ff_expansion_rate
#         self.name = name if name else 'Transformer-Decoder-Block'

#         #LAYERS
#         self.multi_head_self_attention = MultiHeadAttention(self.in_dim,
#                                                             self.key_dim,
#                                                             self.out_dim,
#                                                             self.num_heads,
#                                                             self.drop_out_proba,
#                                                             name = f'{self.name}__MH')
#         self.transformer_block = TransformerBlock(self.in_dim,
#                                                     self.key_dim,
#                                                     self.out_dim,
#                                                     self.num_heads,
#                                                     self.drop_out_proba,
#                                                     self.ff_expansion_rate,
#                                                     name=f'{self.name}__D{i}')


#     def forward(self, x, k, v, src_mask=None, trg_mask=None):
#         q = self.multi_head_self_attention(inputs=x, mask=trg_mask)         # (batch, num_elements, out_dim)
#         outputs = self.transformer_block()
#         feed_forward = self.feed_forward(inputs=attended)                            # (batch, num_elements, out_dim)
#         return feed_forward

# class TransformerDecoderStack(nn.Module):
#     def __init__(self, in_dim,
#                        key_dim,
#                        out_dim,
#                        num_decoders,
#                        num_heads,
#                        drop_out_proba,
#                        ff_expansion_rate,
#                        maximum_position_encoding=None,
#                        name=None):
#         super(TransformerDecoderStack, self).__init__()
#         #PARAMS
#         self.in_dim = in_dim
#         self.key_dim = key_dim
#         self.out_dim = out_dim
#         self.num_decoders = num_decoders
#         self.num_heads = num_heads
#         self.drop_out_proba = drop_out_proba
#         self.ff_expansion_rate = ff_expansion_rate
#         self.maximum_position_encoding = maximum_position_encoding
#         self.name = name if name else 'transformer_decoder_stack'

#         #LAYERS
#         self.transformer_blocks = nn.ModuleList([TransformerBlock(self.in_dim,
#                                                     self.key_dim,
#                                                     self.out_dim,
#                                                     self.num_heads,
#                                                     self.drop_out_proba,
#                                                     self.ff_expansion_rate,
#                                                     name=f'{self.name}__E{i}')
#                                                     for i in range(self.num_decoders)])
#         self.pos_encoder = PositionalEncoding(self.in_dim, self.drop_out_proba, self.maximum_position_encoding)

#     def forward(self, inputs, mask=None):
#         outputs = inputs                                        # (batch, seq_len, emb_size)

#         #POSITION
#         if self.maximum_position_encoding is not None:
#             outputs = self.pos_encoder(outputs)

#         for encoder_layer in self.transformer_blocks:
#             outputs = encoder_layer(inputs=outputs, mask=mask)
#         return outputs                                # (batch, seq_len, out_dim)   <-- USUALLY out_dim = emb_size










from lpd.enums.metric_method import MetricMethod
import torch as T
from lpd.enums.confusion_matrix_based_metric import ConfusionMatrixBasedMetric

class MetricBase(object):
    """
        Args:
        metric_method - from lpd.enums.MetricMethod, use this to dictate how this metric is behaving over the batches,
                        whether its accumulates the MEAN, or the SUM, or taking the LAST value (for example in MetricConfusionMatrixBase)
    """
    def __init__(self, metric_method: MetricMethod):
        self.metric_method = metric_method

    def __call__(self, y_pred: T.Tensor, y_true: T.Tensor):
        raise NotImplementedError('Missing __call__ implementation for metric')


class MetricConfusionMatrixBase(MetricBase):
    def __init__(self, num_classes, labels, predictions_to_classes_convertor, threshold):
        super(MetricConfusionMatrixBase, self).__init__(MetricMethod.LAST)
        self.num_classes = num_classes
        self.labels = labels
        if self.labels and len(self.labels) != num_classes:
            raise ValueError(f'[MetricConfusionMatrixBase] - expecting same number for labels as num_classes, but got num_classes = {num_classes}, and {len(self.labels)} labels')
        self.predictions_to_classes_convertor = predictions_to_classes_convertor
        self.threshold = threshold
        # confusion_matrix_ IS MANAGED INTERNALLY BY TRAINER-STATS VIA _set_confusion_matrix
        self.confusion_matrix_ = None

    def _set_confusion_matrix(self, confusion_matrix):
        """
            INTERNAL USE ONLY!
            the confusion matrix is being handled by TrainerStats, that way there is only one 
            confusion matrix per State (TRAIN/VAL/TEST).
            TrainerStats will inject the most updated confusion matrix here 
        """
        self.confusion_matrix_ = confusion_matrix

    def is_binary(self, y_pred, y_true):
        return self.confusion_matrix_.num_classes == 2

    def get_stats(self, metric: ConfusionMatrixBasedMetric):
        stats = self.confusion_matrix_.get_stats()
        return T.Tensor([stats_per_class[metric] for stats_per_class in stats.values()])

    def get_confusion_matrix(self):
        return self.confusion_matrix_.get_confusion_matrix()
